from typing import List, Union

from benchling_sdk.models import AppSessionMessageCreate, AppSessionMessageStyle


class SessionClosedError(Exception):
    """
    Session Closed Error.

    A session was inoperable because its status in Benchling was terminal.
    """

    pass


class SessionContextClosedError(Exception):
    """
    Session Context Closed Error.

    An operation was attempted using the session context manager after it was closed.
    """

    pass


class InvalidSessionTimeoutError(Exception):
    """
    Invalid Session Timeout Error.

    A session's timeout value was set to an invalid value.
    """

    pass


class MissingAttachedCanvasError(Exception):
    """
    Missing Attached Canvas Error.

    A canvas operation was requested, but a session context has no attached canvas.
    """

    pass


class AppUserFacingError(Exception):
    """
    App User Facing Error.

    Extend this class with custom exceptions you want to be written back to the user as a SessionMessage.

    SessionClosingContextExitHandler will invoke messages() and write them to a user. Callers choosing to
    write their own SessionContextExitHandler may need to replicate this behavior themselves.

    This is useful for control flow where an app wants to terminate with an error state that is resolvable
    by the user.

    For example:

    class InvalidUserInputError(AppUserFacingError):
        pass

    raise InvalidUserInputError("Please enter a number between 1 and 10")

    This would create a message shown to the user like:

    AppSessionMessageCreate("Please enter a number between 1 and 10", style=AppSessionMessageStyle.ERROR)
    """

    _messages: List[str]

    def __init__(self, messages: Union[str, List[str]], *args) -> None:
        """Initialize an AppUserFacingError with one message or a list."""
        self._messages = [messages] if isinstance(messages, str) else messages
        super().__init__(args)

    def messages(self) -> List[AppSessionMessageCreate]:
        """Create a series of AppSessionMessageCreate to write to a Session and displayed to the user."""
        return [
            AppSessionMessageCreate(content=message, style=AppSessionMessageStyle.ERROR)
            for message in self._messages
        ]

    def __str__(self) -> str:
        return "\n".join(self._messages) if self._messages else ""



"""Specialized Exception classes."""
from __future__ import annotations

from abc import ABC, abstractmethod
from dataclasses import dataclass
import json
from json import JSONDecodeError
from typing import Any, Dict, Generic, Iterable, MutableMapping, Optional, Protocol, Type, TypeVar, Union

from benchling_api_client.v2.types import Response

from benchling_sdk.helpers.logging_helpers import default_logger
from benchling_sdk.helpers.serialization_helpers import unset_as_none
from benchling_sdk.models import (
    AsyncTask,
    AsyncTaskStatus,
    BadRequestError,
    BadRequestErrorBulk,
    ConflictError,
    ForbiddenError,
    NotFoundError,
)

logger = default_logger()


@dataclass
class BenchlingError(Exception):
    """
    An error resulting from communicating with the Benchling API.

    This could be an error returned from the API intentionally (e.g., 400 Bad Request) or an
    unexpected transport error (e.g., 502 Bad Gateway)

    The json attribute is present if the API response provided a deserializable JSON body as part of the
    error description. It will be None if the response could not be parsed as JSON.

    The content attribute is any unparsed content returned as part of the response body.

    Instead of extending this class directly, prefer extending ExtendedBenchlingErrorBase.
    """

    status_code: int
    headers: MutableMapping[str, str]
    json: Optional[Dict[str, str]]
    content: Optional[bytes]
    parsed: Union[
        None,
        ForbiddenError,
        NotFoundError,
        BadRequestError,
        BadRequestErrorBulk,
        ConflictError,
    ]

    @classmethod
    def from_response(cls, response: Response) -> BenchlingError:
        """Create a BenchlingError from a generated Response."""
        json_body = _parse_error_body(response.content)
        return cls(
            status_code=response.status_code,
            headers=response.headers,
            json=json_body,
            content=response.content,
            parsed=response.parsed,
        )

    def __str__(self):
        message = self.json if self.json else self.content
        return f"{self.__class__.__name__}(status_code={self.status_code}, message={message})"

    def __hash__(self):
        return self._generate_hash()

    def __eq__(self, other):
        return self._generate_hash() == other._generate_hash()

    def _generate_hash(self):
        return hash(
            (
                self.status_code,
                json.dumps(self.headers, sort_keys=True),
                json.dumps(self.json, sort_keys=True),
                self.content,
                str(self.parsed),
            )
        )


@dataclass
class RegistrationError(Exception):
    """An error relating to Benchling registration."""

    message: Optional[str] = None
    errors: Optional[Dict[Any, Any]] = None
    task_status: Optional[AsyncTaskStatus] = None

    @classmethod
    def from_task(cls, task: AsyncTask) -> RegistrationError:
        """Create a RegistrationError from a failed AsyncTask."""
        task_errors = unset_as_none(lambda: task.errors)
        errors_dict: Dict[Any, Any] = task_errors.to_dict() if task_errors else dict()  # type: ignore
        return cls(
            message=unset_as_none(lambda: task.message), errors=errors_dict, task_status=task.status
        )

    def __str__(self):
        return repr(self)


@dataclass
class WaitForTaskExpiredError(Exception):
    """An error indicating an AsyncTask did not complete in the time allotted for polling."""

    message: str
    task: AsyncTask

    def __str__(self):
        return repr(self)


ExtendedError = TypeVar("ExtendedError", bound="ExtendedBenchlingErrorBase")


class ExtendedBenchlingErrorBase(ABC, BenchlingError):
    """
    Extended Benchling Error.

    Prefer using this class instead of extending BenchlingError directly.

    Forces subclasses to implement useful helpers.
    """

    @classmethod
    @abstractmethod
    def error_matcher(cls: Type[ExtendedError]) -> ResponseErrorMatcher[ExtendedError]:
        """Create an instance of ResponseErrorMatcher matching an exception extending BenchlingError."""
        pass


E = TypeVar("E", bound=BenchlingError)


class ResponseMatcher(Protocol):
    """Callable for checking if a Response matches specified conditions."""

    def __call__(self, response: Response) -> bool:
        """Return True if the provided response matches specified conditions."""
        pass


class ResponseErrorMatcher(Generic[E]):
    """
    Response Error Matcher.

    Raise a more specific Benchling Error if the HTTP response matches specific conditions.
    """

    _error_type: Type[E]
    _matcher: ResponseMatcher

    def __init__(self, error_type: Type[E], matcher: ResponseMatcher):
        """Init Response Error Matcher."""
        self._error_type = error_type
        self._matcher = matcher

    def raise_on_match(self, response: Response) -> None:
        """Raise error_type if the response matches some defined condition."""
        if self._matcher(response):
            raise self._error_type.from_response(response)


class AppSessionClosedError(ExtendedBenchlingErrorBase):
    """An error indicating a Benchling App session was already closed when attempting an API update."""

    _ERROR_TEXT: str = "Session has been closed and cannot be updated"

    @classmethod
    def error_matcher(cls) -> ResponseErrorMatcher[AppSessionClosedError]:
        """Create an instance of ResponseErrorMatcher matching AppSessionClosedError."""

        def _matcher(response: Response) -> bool:
            return _match_in_response_message(response, cls._ERROR_TEXT, status_code=400)

        return ResponseErrorMatcher(cls, _matcher)


@dataclass
class InvalidDataFrameError(Exception):
    """A general error related to Benchling DataFrames."""

    message: str

    def __str__(self):
        return repr(self)


@dataclass
class DataFrameInProgressError(Exception):
    """An error for Benchling data frames for unavailable data operations on a DataFrame in progress."""

    message: str

    def __str__(self):
        return repr(self)


def raise_for_status(
    response: Response, error_matchers: Optional[Iterable[ResponseErrorMatcher[E]]] = None
) -> Response:
    """
    Evaluate a Response for a successful HTTP status code or raise a BenchlingError.

    Custom error_matchers may be passed to raise more specific types of BenchlingError in a few specialized cases.

    If multiple matchers are specified, it's recommended that the caller provide an ordered Iterable
    such as List, with the more specific matchers first in the collection. Specifying a broader exception
    earlier in the collection, with a more narrow exception later, may cause the broader exception to be
    returned first.
    """
    logger.info("Status: %s", response.status_code)
    logger.debug(response)
    if response.status_code < 200 or response.status_code >= 300:
        # Check for specific error matches
        if error_matchers:
            for check_matcher in error_matchers:
                check_matcher.raise_on_match(response)
        # Raise a generalized error
        raise BenchlingError.from_response(response)
    return response


def _parse_error_body(content: bytes) -> Optional[Dict[str, str]]:
    if content:
        string_content = content.decode("utf-8")
        # In case the response is not a serialized JSON dict (e.g. a gateway error)
        try:
            return json.loads(string_content)
        except JSONDecodeError:
            # Some responses may not be JSON. Just catch the error
            return None
    return None


def _match_in_response_message(
    response: Response, match_string: str, status_code: Optional[int] = None
) -> bool:
    if status_code is not None and status_code != response.status_code:
        return False
    json_body = _parse_error_body(response.content)
    error_message = _error_message(json_body) if json_body is not None else None
    return error_message is not None and match_string in error_message


def _error_message(body: Dict[str, Any]) -> Optional[str]:
    error = body.get("error", {})
    # Needed for MyPy
    if isinstance(error, dict):
        return error.get("message")
    return None



from typing import Any, Dict, Iterable, List, Optional, Union

from benchling_api_client.v2.stable.api.oligos import (
    archive_oligos,
    bulk_create_oligos,
    bulk_get_oligos,
    create_oligo,
    get_oligo,
    list_oligos,
    unarchive_oligos,
    update_oligo,
)
from benchling_api_client.v2.types import Response

from benchling_sdk.errors import raise_for_status
from benchling_sdk.helpers.constants import _translate_to_string_enum
from benchling_sdk.helpers.decorators import api_method
from benchling_sdk.helpers.logging_helpers import check_for_csv_bug_fix, log_deprecation
from benchling_sdk.helpers.pagination_helpers import NextToken, PageIterator
from benchling_sdk.helpers.response_helpers import model_from_detailed
from benchling_sdk.helpers.serialization_helpers import (
    none_as_unset,
    optional_array_query_param,
    schema_fields_query_param,
)
from benchling_sdk.helpers.task_helpers import TaskHelper
from benchling_sdk.models import (
    BulkCreateDnaOligosAsyncTaskResponse,
    DnaOligo,
    EntityArchiveReason,
    ListOligosSort,
    Oligo,
    OligoCreate,
    OligosArchivalChange,
    OligosArchive,
    OligosBulkCreateRequest,
    OligosPaginatedList,
    OligosUnarchive,
    OligoUpdate,
    RnaOligo,
)
from benchling_sdk.services.v2.base_service import BaseService


class OligoService(BaseService):
    """
    Oligos (Deprecated).

    Oligos are short linear DNA sequences that can be attached as primers to full DNA sequences. Just like other
    entities, they support schemas, tags, and aliases.

    Please migrate to the corresponding DNA/RNA Oligos endpoints.

    See https://benchling.com/api/reference#/Oligos
    See https://benchling.com/api/reference#/DNA%20Oligos
    See https://benchling.com/api/reference#/RNA%20Oligos
    """

    @api_method
    def get_by_id(self, oligo_id: str, returning: Optional[Iterable[str]] = None) -> DnaOligo:
        """
        Get an Oligo by ID (Deprecated).

        Please migrate to the corresponding DNA/RNA Oligos endpoints.

        See https://benchling.com/api/reference#/Oligos/getOligo
        See https://benchling.com/api/reference#/DNA%20Oligos/getDNAOligo
        See https://benchling.com/api/reference#/RNA%20Oligos/getRNAOligo
        """
        log_deprecation("oligos.get_by_id", "dna_oligos.get_by_id or rna_oligos.get_by_id")
        returning_string = optional_array_query_param(returning)
        response = get_oligo.sync_detailed(
            client=self.client, oligo_id=oligo_id, returning=none_as_unset(returning_string)
        )
        return model_from_detailed(response)

    @api_method
    def _oligos_page(
        self,
        modified_at: Optional[str] = None,
        name: Optional[str] = None,
        bases: Optional[str] = None,
        folder_id: Optional[str] = None,
        mentioned_in: Optional[List[str]] = None,
        project_id: Optional[str] = None,
        registry_id: Optional[str] = None,
        schema_id: Optional[str] = None,
        archive_reason: Optional[str] = None,
        mentions: Optional[List[str]] = None,
        sort: Optional[ListOligosSort] = None,
        ids: Optional[Iterable[str]] = None,
        entity_registry_ids_any_of: Optional[Iterable[str]] = None,
        names_any_of: Optional[Iterable[str]] = None,
        names_any_of_case_sensitive: Optional[Iterable[str]] = None,
        schema_fields: Optional[Dict[str, Any]] = None,
        creator_ids: Optional[Iterable[str]] = None,
        page_size: Optional[int] = None,
        next_token: Optional[NextToken] = None,
        returning: Optional[Iterable[str]] = None,
    ) -> Response[OligosPaginatedList]:
        """Deprecate in favor of dna_oligos._dna_oligos_page or rna_oligos._rna_oligos_page."""
        log_deprecation("oligos._oligos_page", "dna_oligos._dna_oligos_page or rna_oligos._rna_oligos_page")
        response = list_oligos.sync_detailed(
            client=self.client,
            modified_at=none_as_unset(modified_at),
            name=none_as_unset(name),
            bases=none_as_unset(bases),
            folder_id=none_as_unset(folder_id),
            mentioned_in=none_as_unset(optional_array_query_param(mentioned_in)),
            project_id=none_as_unset(project_id),
            registry_id=none_as_unset(registry_id),
            schema_id=none_as_unset(schema_id),
            archive_reason=none_as_unset(archive_reason),
            mentions=none_as_unset(optional_array_query_param(mentions)),
            sort=none_as_unset(sort),
            ids=none_as_unset(optional_array_query_param(ids)),
            entity_registry_idsany_of=none_as_unset(optional_array_query_param(entity_registry_ids_any_of)),
            namesany_of=none_as_unset(optional_array_query_param(names_any_of)),
            namesany_ofcase_sensitive=none_as_unset(optional_array_query_param(names_any_of_case_sensitive)),
            schema_fields=none_as_unset(schema_fields_query_param(schema_fields)),
            creator_ids=none_as_unset(optional_array_query_param(creator_ids)),
            page_size=none_as_unset(page_size),
            next_token=none_as_unset(next_token),
            returning=none_as_unset(optional_array_query_param(returning)),
        )
        raise_for_status(response)
        return response  # type: ignore

    def list(
        self,
        modified_at: Optional[str] = None,
        name: Optional[str] = None,
        bases: Optional[str] = None,
        folder_id: Optional[str] = None,
        mentioned_in: Optional[List[str]] = None,
        project_id: Optional[str] = None,
        registry_id: Optional[str] = None,
        schema_id: Optional[str] = None,
        archive_reason: Optional[str] = None,
        mentions: Optional[List[str]] = None,
        ids: Optional[Iterable[str]] = None,
        entity_registry_ids_any_of: Optional[Iterable[str]] = None,
        names_any_of: Optional[Iterable[str]] = None,
        names_any_of_case_sensitive: Optional[Iterable[str]] = None,
        schema_fields: Optional[Dict[str, Any]] = None,
        creator_ids: Optional[Iterable[str]] = None,
        sort: Optional[Union[str, ListOligosSort]] = None,
        page_size: Optional[int] = None,
        returning: Optional[Iterable[str]] = None,
    ) -> PageIterator[Oligo]:
        """
        List Oligos (Deprecated).

        Please migrate to the corresponding DNA/RNA Oligos endpoints.

        See https://benchling.com/api/reference#/Oligos/listOligos
        See https://benchling.com/api/reference#/DNA%20Oligos/listDNAOligos
        See https://benchling.com/api/reference#/RNA%20Oligos/listRNAOligos
        """
        log_deprecation("oligos.list", "dna_oligos.list or rna_oligos.list")
        check_for_csv_bug_fix("mentioned_in", mentioned_in)
        check_for_csv_bug_fix("mentions", mentions)

        def api_call(next_token: NextToken) -> Response[OligosPaginatedList]:
            return self._oligos_page(
                modified_at=modified_at,
                name=name,
                bases=bases,
                folder_id=folder_id,
                mentioned_in=mentioned_in,
                project_id=project_id,
                registry_id=registry_id,
                schema_id=schema_id,
                archive_reason=archive_reason,
                mentions=mentions,
                ids=ids,
                entity_registry_ids_any_of=entity_registry_ids_any_of,
                names_any_of=names_any_of,
                names_any_of_case_sensitive=names_any_of_case_sensitive,
                schema_fields=schema_fields,
                creator_ids=creator_ids,
                sort=_translate_to_string_enum(ListOligosSort, sort),
                page_size=page_size,
                next_token=next_token,
                returning=returning,
            )

        def results_extractor(body: OligosPaginatedList) -> Optional[List[Oligo]]:
            return body.oligos

        return PageIterator(api_call, results_extractor)

    @api_method
    def create(self, oligo: OligoCreate) -> DnaOligo:
        """
        Create an Oligo (Deprecated).

        Please migrate to the corresponding DNA/RNA Oligos endpoints.

        See https://benchling.com/api/reference#/Oligos/createOligo
        See https://benchling.com/api/reference#/DNA%20Oligos/createDNAOligo
        See https://benchling.com/api/reference#/RNA%20Oligos/createRNAOligo
        """
        log_deprecation("oligos.create", "dna_oligos.create or rna_oligos.create")
        response = create_oligo.sync_detailed(client=self.client, json_body=oligo)
        return model_from_detailed(response)

    @api_method
    def update(self, oligo_id: str, oligo: OligoUpdate) -> DnaOligo:
        """
        Update an Oligo (Deprecated).

        Please migrate to the corresponding DNA/RNA Oligos endpoints.

        See https://benchling.com/api/reference#/Oligos/updateOligo
        See https://benchling.com/api/reference#/DNA%20Oligos/updateDNAOligo
        See https://benchling.com/api/reference#/RNA%20Oligos/updateRNAOligo
        """
        log_deprecation("oligos.update", "dna_oligos.update or rna_oligos.update")
        response = update_oligo.sync_detailed(client=self.client, oligo_id=oligo_id, json_body=oligo)
        return model_from_detailed(response)

    @api_method
    def archive(self, oligo_ids: Iterable[str], reason: EntityArchiveReason) -> OligosArchivalChange:
        """
        Archive Oligos (Deprecated).

        Please migrate to the corresponding DNA/RNA Oligos endpoints.

        See https://benchling.com/api/reference#/Oligos/archiveOligos
        See https://benchling.com/api/reference#/DNA%20Oligos/archiveDNAOligos
        See https://benchling.com/api/reference#/RNA%20Oligos/archiveRNAOligos
        """
        log_deprecation("oligos.archive", "dna_oligos.archive or rna_oligos.archive")
        archive_request = OligosArchive(reason=reason, oligo_ids=list(oligo_ids))
        response = archive_oligos.sync_detailed(client=self.client, json_body=archive_request)
        return model_from_detailed(response)

    @api_method
    def unarchive(self, oligo_ids: Iterable[str]) -> OligosArchivalChange:
        """
        Unarchive Oligos (Deprecated).

        Please migrate to the corresponding DNA/RNA Oligos endpoints.

        See https://benchling.com/api/reference#/Oligos/unarchiveOligos
        See https://benchling.com/api/reference#/DNA%20Oligos/unarchiveDNAOligos
        See https://benchling.com/api/reference#/RNA%20Oligos/unarchiveRNAOligos
        """
        log_deprecation("oligos.unarchive", "dna_oligos.unarchive or rna_oligos.unarchive")
        unarchive_request = OligosUnarchive(oligo_ids=list(oligo_ids))
        response = unarchive_oligos.sync_detailed(client=self.client, json_body=unarchive_request)
        return model_from_detailed(response)

    @api_method
    def bulk_get(
        self, oligo_ids: Iterable[str], returning: Optional[Iterable[str]] = None
    ) -> Optional[List[Union[DnaOligo, RnaOligo]]]:
        """
        Bulk get Oligos.

        See https://benchling.com/api/reference#/Oligos/bulkGetOligos
        """
        oligo_id_string = ",".join(oligo_ids)
        returning_string = optional_array_query_param(returning)
        response = bulk_get_oligos.sync_detailed(
            client=self.client, oligo_ids=oligo_id_string, returning=none_as_unset(returning_string)
        )
        oligos_results = model_from_detailed(response)
        return oligos_results.oligos

    @api_method
    def bulk_create(self, oligos: Iterable[OligoCreate]) -> TaskHelper[BulkCreateDnaOligosAsyncTaskResponse]:
        """
        Bulk create DNA Oligos (Deprecated).

        Please migrate to the corresponding DNA/RNA Oligos endpoints.

        See https://benchling.com/api/reference#/Oligos/bulkCreateOligos
        See https://benchling.com/api/reference#/DNA%20Oligos/bulkCreateDNAOligos
        See https://benchling.com/api/reference#/RNA%20Oligos/bulkCreateRNAOligos
        """
        log_deprecation("oligos.bulk_create", "dna_oligos.bulk_create or rna_oligos.bulk_create")
        body = OligosBulkCreateRequest(list(oligos))
        response = bulk_create_oligos.sync_detailed(client=self.client, json_body=body)
        return self._task_helper_from_response(response, BulkCreateDnaOligosAsyncTaskResponse)



from __future__ import annotations

from typing import Protocol

from benchling_api_client.v2.benchling_client import AuthorizationMethod


class BearerTokenProvider(Protocol):
    """A callable function for providing a bearer token."""

    def __call__(self, base_url: str) -> str:
        """Return a bearer token to be used with an HTTP Authorization header."""
        pass


class BearerTokenAuth(AuthorizationMethod):
    """
    Bearer Token Authorization.

    Use in combination with the Benchling() client constructor to be authorized with Bearer Token Authorization
    """

    _token_function: BearerTokenProvider

    def __init__(self, token_function: BearerTokenProvider) -> None:
        """Init BearerTokenAuth with a callable BearerTokenProvider for providing bearer tokens."""
        super().__init__()
        self._token_function = token_function

    @classmethod
    def from_token(cls, bearer_token: str) -> BearerTokenAuth:
        """Create a BearerTokenAuth that always uses a static bearer token string."""
        return cls(lambda base_url: bearer_token)

    def get_authorization_header(self, base_url: str) -> str:
        """Get content for a HTTP Authorization header."""
        return f"Bearer {self._token_function(base_url)}"
